# -*- coding: utf-8 -*-
#
# Copyright © 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluža <jkaluza@redhat.com>
#            Igor Gnatenko <ignatenko@redhat.com>

import collections
import pytest

from module_build_service.mmd_resolver import MMDResolver
from module_build_service import Modulemd
from module_build_service import glib


class TestMMDResolver:

    def setup_method(self, test_method):
        self.mmd_resolver = MMDResolver()

    def teardown_method(self, test_method):
        pass

    @staticmethod
    def _make_mmd(nsvc, requires, xmd_buildrequires=None, virtual_streams=None):
        name, stream, version = nsvc.split(":", 2)
        mmd = Modulemd.Module()
        mmd.set_mdversion(2)
        mmd.set_name(name)
        mmd.set_stream(stream)
        mmd.set_summary("foo")
        mmd.set_description("foo")
        licenses = Modulemd.SimpleSet()
        licenses.add("GPL")
        mmd.set_module_licenses(licenses)

        xmd = glib.from_variant_dict(mmd.get_xmd())
        xmd["mbs"] = {}
        xmd["mbs"]["buildrequires"] = {}
        if xmd_buildrequires:
            for ns in xmd_buildrequires:
                n, s = ns.split(":")
                xmd["mbs"]["buildrequires"][n] = {"stream": s}
        if virtual_streams:
            xmd["mbs"]["virtual_streams"] = virtual_streams
        mmd.set_xmd(glib.dict_values(xmd))

        if ":" in version:
            version, context = version.split(":")
            mmd.set_context(context)
            add_requires = Modulemd.Dependencies.add_requires
        else:
            add_requires = Modulemd.Dependencies.add_buildrequires
        mmd.set_version(int(version))

        if not isinstance(requires, list):
            requires = [requires]
        else:
            requires = requires

        deps_list = []
        for reqs in requires:
            deps = Modulemd.Dependencies()
            for req_name, req_streams in reqs.items():
                add_requires(deps, req_name, req_streams)
            deps_list.append(deps)
        mmd.set_dependencies(deps_list)

        return mmd

    @pytest.mark.parametrize(
        "deps, expected", (
            ([], "None"),
            ([{"x": []}], "module(x)"),
            ([{"x": ["1"]}], "(module(x) with module(x:1))"),
            ([{"x": ["-1"]}], "(module(x) without module(x:1))"),
            ([{"x": ["1", "2"]}], "(module(x) with (module(x:1) or module(x:2)))"),
            ([{"x": ["-1", "2"]}], "(module(x) with module(x:2))"),
            ([{"x": [], "y": []}], "(module(x) and module(y))"),
            ([{"x": []}, {"y": []}], "(module(x) or module(y))"),
        )
    )
    def test_deps2reqs(self, deps, expected):
        # Sort by keys here to avoid unordered dicts
        deps = [collections.OrderedDict(sorted(dep.items())) for dep in deps]
        reqs = self.mmd_resolver._deps2reqs(deps)
        assert str(reqs) == expected

    @pytest.mark.parametrize(
        "buildrequires, expected", (
            ({"platform": []}, [
                [["platform:f28:0:c0:x86_64"],
                 ["platform:f29:0:c0:x86_64"]],
            ]),
            ({"platform": ["f28"]}, [
                [["platform:f28:0:c0:x86_64"]],
            ]),
            ({"platform": ["-f28"]}, [
                [["platform:f29:0:c0:x86_64"]],
            ]),
            ({"gtk": [], "qt": []}, [
                [["gtk:3:0:c8:x86_64", "qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                 ["gtk:4:0:c8:x86_64", "qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                 ["gtk:3:0:c8:x86_64", "qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                 ["gtk:4:0:c8:x86_64", "qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"]],
            ]),
            ({"gtk": [], "qt": [], "platform": []}, [
                [["gtk:3:0:c8:x86_64", "qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                 ["gtk:4:0:c8:x86_64", "qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                 ["gtk:3:0:c8:x86_64", "qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                 ["gtk:4:0:c8:x86_64", "qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                 ["gtk:3:0:c9:x86_64", "qt:4:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                 ["gtk:4:0:c9:x86_64", "qt:4:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                 ["gtk:3:0:c9:x86_64", "qt:5:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                 ["gtk:4:0:c9:x86_64", "qt:5:0:c9:x86_64", "platform:f29:0:c0:x86_64"]],
            ]),
            ([{"qt": [], "platform": ["f28"]},
              {"gtk": [], "platform": ["-f28"]}], [
                [["qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                 ["qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"]],
                [["gtk:3:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                 ["gtk:4:0:c9:x86_64", "platform:f29:0:c0:x86_64"]],
            ]),
            ({"mess": []}, [
                [["mess:1:0:c0:x86_64", "gtk:3:0:c8:x86_64", "platform:f28:0:c0:x86_64"]],
            ]),
            ({"mess": [], "platform": []}, [
                [["mess:1:0:c0:x86_64", "gtk:3:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                 ["mess:1:0:c0:x86_64", "gtk:4:0:c9:x86_64", "platform:f29:0:c0:x86_64"]],
            ]),
        )
    )
    def test_solve(self, buildrequires, expected):
        modules = (
            ("platform:f28:0:c0", {}),
            ("platform:f29:0:c0", {}),
            ("gtk:3:0:c8", {"platform": ["f28"]}),
            ("gtk:3:0:c9", {"platform": ["f29"]}),
            ("gtk:4:0:c8", {"platform": ["f28"]}),
            ("gtk:4:0:c9", {"platform": ["f29"]}),
            ("qt:4:0:c8", {"platform": ["f28"]}),
            ("qt:4:0:c9", {"platform": ["f29"]}),
            ("qt:5:0:c8", {"platform": ["f28"]}),
            ("qt:5:0:c9", {"platform": ["f29"]}),
            ("mess:1:0:c0", [{"gtk": ["3"], "platform": ["f28"]},
                             {"gtk": ["4"], "platform": ["-f28"]}]),
        )
        for n, req in modules:
            self.mmd_resolver.add_modules(self._make_mmd(n, req))

        app = self._make_mmd("app:1:0", buildrequires)
        expanded = self.mmd_resolver.solve(app)

        expected = set(frozenset(["app:1:0:%d:src" % c] + e)
                       for c, exp in enumerate(expected)
                       for e in exp)

        assert expanded == expected

    @pytest.mark.parametrize(
        "buildrequires, xmd_buildrequires, expected", (
            # BR all platform streams -> build for all platform streams.
            ({"platform": []}, {}, [
                [["platform:el8.2.0:0:c0:x86_64"],
                 ["platform:el8.1.0:0:c0:x86_64"],
                 ["platform:el8.0.0:0:c0:x86_64"],
                 ["platform:el7.6.0:0:c0:x86_64"]],
            ]),
            # BR "el8" platform stream -> build for all el8 platform streams.
            ({"platform": ["el8"]}, {}, [
                [["platform:el8.2.0:0:c0:x86_64"],
                 ["platform:el8.1.0:0:c0:x86_64"],
                 ["platform:el8.0.0:0:c0:x86_64"]],
            ]),
            # BR "el8.1.0" platfrom stream -> build just for el8.1.0.
            ({"platform": ["el8"]}, ["platform:el8.1.0"], [
                [["platform:el8.1.0:0:c0:x86_64"]],
            ]),
            # BR platform:el8.1.0 and gtk:3, which is not built against el8.1.0,
            # but it is built only against el8.0.0 -> cherry-pick gtk:3 from el8.0.0
            # and build once against platform:el8.1.0.
            ({"platform": ["el8"], "gtk": ["3"]}, ["platform:el8.1.0"], [
                [["platform:el8.1.0:0:c0:x86_64", "gtk:3:0:c8:x86_64", ]],
            ]),
            # BR platform:el8.2.0 and gtk:3, this time gtk:3 build against el8.2.0 exists
            # -> use both platform and gtk from el8.2.0 and build once.
            ({"platform": ["el8"], "gtk": ["3"]}, ["platform:el8.2.0"], [
                [["platform:el8.2.0:0:c0:x86_64", "gtk:3:1:c8:x86_64", ]],
            ]),
            # BR platform:el8.2.0 and mess:1 which is built against platform:el8.1.0 and
            # requires gtk:3 which is built against platform:el8.2.0 and platform:el8.0.0
            # -> Use platform:el8.2.0 and
            # -> cherry-pick mess:1 from el8.1.0 and
            # -> use gtk:3:1 from el8.2.0.
            ({"platform": ["el8"], "mess": ["1"]}, ["platform:el8.2.0"], [
                [["platform:el8.2.0:0:c0:x86_64", "mess:1:0:c0:x86_64", "gtk:3:1:c8:x86_64", ]],
            ]),
            # BR platform:el8.1.0 and mess:1 which is built against platform:el8.1.0 and
            # requires gtk:3 which is built against platform:el8.2.0 and platform:el8.0.0
            # -> Use platform:el8.1.0 and
            # -> Used mess:1 from el8.1.0 and
            # -> cherry-pick gtk:3:0 from el8.0.0.
            ({"platform": ["el8"], "mess": ["1"]}, ["platform:el8.1.0"], [
                [["platform:el8.1.0:0:c0:x86_64", "mess:1:0:c0:x86_64", "gtk:3:0:c8:x86_64", ]],
            ]),
            # BR platform:el8.0.0 and mess:1 which is built against platform:el8.1.0 and
            # requires gtk:3 which is built against platform:el8.2.0 and platform:el8.0.0
            # -> No valid combination, because mess:1 is only available in el8.1.0 and later.
            ({"platform": ["el8"], "mess": ["1"]}, ["platform:el8.0.0"], []),
            # This is undefined... it might build just once against latest platform or
            # against all the platforms... we don't know
            # ({"platform": ["el8"], "gtk": ["3"]}, {}, [
            #     [["platform:el8.2.0:0:c0:x86_64", "gtk:3:1:c8:x86_64"]],
            # ]),
        )
    )
    def test_solve_virtual_streams(self, buildrequires, xmd_buildrequires, expected):
        modules = (
            # (nsvc, buildrequires, expanded_buildrequires, virtual_streams)
            ("platform:el8.0.0:0:c0", {}, {}, ["el8"]),
            ("platform:el8.1.0:0:c0", {}, {}, ["el8"]),
            ("platform:el8.2.0:0:c0", {}, {}, ["el8"]),
            ("platform:el7.6.0:0:c0", {}, {}, ["el7"]),
            ("gtk:3:0:c8", {"platform": ["el8"]}, {"platform:el8.0.0"}, None),
            ("gtk:3:1:c8", {"platform": ["el8"]}, {"platform:el8.2.0"}, None),
            ("mess:1:0:c0", [{"gtk": ["3"], "platform": ["el8"]}], {"platform:el8.1.0"}, None),
        )
        for n, req, xmd_br, virtual_streams in modules:
            self.mmd_resolver.add_modules(self._make_mmd(
                n, req, xmd_br, virtual_streams))

        app = self._make_mmd("app:1:0", buildrequires, xmd_buildrequires)
        if not expected:
            with pytest.raises(RuntimeError):
                self.mmd_resolver.solve(app)
            return
        else:
            expanded = self.mmd_resolver.solve(app)

        expected = set(frozenset(["app:1:0:%d:src" % c] + e)
                       for c, exp in enumerate(expected)
                       for e in exp)

        assert expanded == expected

    @pytest.mark.parametrize('app_buildrequires, modules, err_msg_regex', (
        # app --br--> gtk:1 --req--> bar:1* ---req---> platform:f29
        #    \--br--> foo:1 --req--> bar:2* ---req--/
        (
            {'gtk': '1', 'foo': '1'},
            (
                ('platform:f29:0:c0', {}),
                ('gtk:1:1:c01', {'bar': ['1']}),
                ('bar:1:0:c02', {'platform': ['f29']}),
                ('foo:1:1:c03', {'bar': ['2']}),
                ('bar:2:0:c04', {'platform': ['f29']}),
            ),
            'bar:1:0:c02 and bar:2:0:c04',
        ),
        # app --br--> gtk:1 --req--> bar:1* ----------req----------> platform:f29
        #    \--br--> foo:1 --req--> baz:1 --req--> bar:2* --req--/
        (
            {'gtk': '1', 'foo': '1'},
            (
                ('platform:f29:0:c0', {}),

                ('gtk:1:1:c01', {'bar': ['1']}),
                ('bar:1:0:c02', {'platform': ['f29']}),

                ('foo:1:1:c03', {'baz': ['1']}),
                ('baz:1:1:c04', {'bar': ['2']}),
                ('bar:2:0:c05', {'platform': ['f29']}),
            ),
            'bar:1:0:c02 and bar:2:0:c05',
        ),
        # Test multiple conflicts pairs are detected.
        # app --br--> gtk:1 --req--> bar:1* ---------req-----------\
        #    \--br--> foo:1 --req--> baz:1 --req--> bar:2* ---req---> platform:f29
        #    \--br--> pkga:1 --req--> perl:5' -------req-----------/
        #    \--br--> pkgb:1 --req--> perl:6' -------req-----------/
        (
            {'gtk': '1', 'foo': '1', 'pkga': '1', 'pkgb': '1'},
            (
                ('platform:f29:0:c0', {}),

                ('gtk:1:1:c01', {'bar': ['1']}),
                ('bar:1:0:c02', {'platform': ['f29']}),

                ('foo:1:1:c03', {'baz': ['1']}),
                ('baz:1:1:c04', {'bar': ['2']}),
                ('bar:2:0:c05', {'platform': ['f29']}),

                ('pkga:1:0:c06', {'perl': ['5']}),
                ('perl:5:0:c07', {'platform': ['f29']}),

                ('pkgb:1:0:c08', {'perl': ['6']}),
                ('perl:6:0:c09', {'platform': ['f29']}),
            ),
            # MMD Resolver should still catch a conflict
            'bar:1:0:c02 and bar:2:0:c05',
        ),
    ))
    def test_solve_stream_conflicts(self, app_buildrequires, modules, err_msg_regex):
        for n, req in modules:
            self.mmd_resolver.add_modules(self._make_mmd(n, req))

        app = self._make_mmd("app:1:0", app_buildrequires)

        with pytest.raises(RuntimeError, match=err_msg_regex):
            self.mmd_resolver.solve(app)



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Stanislav Ochotnicky <sochotnicky@redhat.com>
#            Jan Kaluza <jkaluza@redhat.com>

import pytest
import json

import os
from os import path

import module_build_service.messaging
import module_build_service.scheduler.handlers.repos # noqa
from module_build_service import models, conf, build_logs, Modulemd, glib

from mock import patch, Mock, call, mock_open
import kobo.rpmlib

from tests import init_data
from tests.test_views.test_views import FakeSCM

from module_build_service.builder.KojiContentGenerator import KojiContentGenerator

GET_USER_RV = {
    "id": 3686,
    "krb_principal": "mszyslak@FEDORAPROJECT.ORG",
    "name": "Moe Szyslak",
    "status": 0,
    "usertype": 0
}


class TestBuild:

    def setup_method(self, test_method):
        init_data(1, contexts=True)
        module = models.ModuleBuild.query.filter_by(id=2).one()
        module.cg_build_koji_tag = "f27-module-candidate"
        self.cg = KojiContentGenerator(module, conf)

        self.p_read_config = patch('koji.read_config', return_value={
            'authtype': 'kerberos',
            'timeout': 60,
            'server': 'http://koji.example.com/'
        })
        self.mock_read_config = self.p_read_config.start()

        # Ensure that there is no build log from other tests
        try:
            file_path = build_logs.path(self.cg.module)
            os.remove(file_path)
        except OSError:
            pass

    def teardown_method(self, test_method):
        self.p_read_config.stop()

        # Necessary to restart the twisted reactor for the next test.
        import sys
        del sys.modules['twisted.internet.reactor']
        del sys.modules['moksha.hub.reactor']
        del sys.modules['moksha.hub']
        import moksha.hub.reactor # noqa
        try:
            file_path = build_logs.path(self.cg.module)
            os.remove(file_path)
        except OSError:
            pass

    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    @patch("subprocess.Popen")
    @patch("subprocess.check_output", return_value='1.4')
    @patch("pkg_resources.get_distribution")
    @patch("platform.linux_distribution")
    @patch("platform.machine")
    @patch(("module_build_service.builder.KojiContentGenerator.KojiContentGenerator."
           "_koji_rpms_in_tag"))
    @pytest.mark.parametrize("devel", (False, True))
    def test_get_generator_json(self, rpms_in_tag, machine, distro, pkg_res, coutput, popen,
                                ClientSession, devel):
        """ Test generation of content generator json """
        koji_session = ClientSession.return_value
        koji_session.getUser.return_value = GET_USER_RV
        koji_session.getTag.return_value = {"arches": ""}
        distro.return_value = ("Fedora", "25", "Twenty Five")
        machine.return_value = "i686"
        pkg_res.return_value = Mock()
        pkg_res.return_value.version = "current-tested-version"
        rpm_mock = Mock()
        rpm_out = "rpm-name;1.0;r1;x86_64;(none);sigmd5:1;sigpgp:p;siggpg:g\n" \
                  "rpm-name-2;2.0;r2;i686;1;sigmd5:2;sigpgp:p2;siggpg:g2"
        attrs = {'communicate.return_value': (rpm_out, 'error'),
                 'wait.return_value': 0}
        rpm_mock.configure_mock(**attrs)
        popen.return_value = rpm_mock

        tests_dir = path.abspath(path.dirname(__file__))
        rpm_in_tag_path = path.join(tests_dir,
                                    "test_get_generator_json_rpms_in_tag.json")
        with open(rpm_in_tag_path) as rpms_in_tag_file:
            rpms_in_tag.return_value = json.load(rpms_in_tag_file)

        expected_output_path = path.join(tests_dir,
                                         "test_get_generator_json_expected_output_with_log.json")
        with open(expected_output_path) as expected_output_file:
            expected_output = json.load(expected_output_file)

        # create the build.log
        build_logs.start(self.cg.module)
        build_logs.stop(self.cg.module)

        self.cg.devel = devel
        self.cg._load_koji_tag(koji_session)
        file_dir = self.cg._prepare_file_directory()
        ret = self.cg._get_content_generator_metadata(file_dir)
        rpms_in_tag.assert_called_once()
        if not devel:
            assert expected_output == ret
        else:
            # For devel, only check that the name has -devel suffix.
            assert ret["build"]["name"] == "nginx-devel"
            assert ret["build"]["extra"]["typeinfo"]["module"]["name"] == "nginx-devel"

        # Ensure an anonymous Koji session works
        koji_session.krb_login.assert_not_called()

    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    @patch("subprocess.Popen")
    @patch("subprocess.check_output", return_value='1.4')
    @patch("pkg_resources.get_distribution")
    @patch("platform.linux_distribution")
    @patch("platform.machine")
    @patch(("module_build_service.builder.KojiContentGenerator.KojiContentGenerator."
           "_koji_rpms_in_tag"))
    def test_get_generator_json_no_log(self, rpms_in_tag, machine, distro, pkg_res, coutput, popen,
                                       ClientSession):
        """ Test generation of content generator json """
        koji_session = ClientSession.return_value
        koji_session.getUser.return_value = GET_USER_RV
        koji_session.getTag.return_value = {"arches": ""}
        distro.return_value = ("Fedora", "25", "Twenty Five")
        machine.return_value = "i686"
        pkg_res.return_value = Mock()
        pkg_res.return_value.version = "current-tested-version"
        rpm_mock = Mock()
        rpm_out = "rpm-name;1.0;r1;x86_64;(none);sigmd5:1;sigpgp:p;siggpg:g\n" \
                  "rpm-name-2;2.0;r2;i686;1;sigmd5:2;sigpgp:p2;siggpg:g2"
        attrs = {'communicate.return_value': (rpm_out, 'error'),
                 'wait.return_value': 0}
        rpm_mock.configure_mock(**attrs)
        popen.return_value = rpm_mock

        tests_dir = path.abspath(path.dirname(__file__))
        rpm_in_tag_path = path.join(tests_dir,
                                    "test_get_generator_json_rpms_in_tag.json")
        with open(rpm_in_tag_path) as rpms_in_tag_file:
            rpms_in_tag.return_value = json.load(rpms_in_tag_file)

        expected_output_path = path.join(tests_dir,
                                         "test_get_generator_json_expected_output.json")
        with open(expected_output_path) as expected_output_file:
            expected_output = json.load(expected_output_file)
        self.cg._load_koji_tag(koji_session)
        file_dir = self.cg._prepare_file_directory()
        ret = self.cg._get_content_generator_metadata(file_dir)
        rpms_in_tag.assert_called_once()
        assert expected_output == ret

        # Anonymous koji session should work well.
        koji_session.krb_login.assert_not_called()

    def test_prepare_file_directory(self):
        """ Test preparation of directory with output files """
        dir_path = self.cg._prepare_file_directory()
        with open(path.join(dir_path, "modulemd.txt")) as mmd:
            assert len(mmd.read()) == 1134

    def test_prepare_file_directory_per_arch_mmds(self):
        """ Test preparation of directory with output files """
        self.cg.arches = ["x86_64", "i686"]
        dir_path = self.cg._prepare_file_directory()
        with open(path.join(dir_path, "modulemd.txt")) as mmd:
            assert len(mmd.read()) == 1134

        with open(path.join(dir_path, "modulemd.x86_64.txt")) as mmd:
            assert len(mmd.read()) == 257

        with open(path.join(dir_path, "modulemd.i686.txt")) as mmd:
            assert len(mmd.read()) == 255

    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    def test_tag_cg_build(self, ClientSession):
        """ Test that the CG build is tagged. """
        koji_session = ClientSession.return_value
        koji_session.getUser.return_value = GET_USER_RV
        koji_session.getTag.return_value = {'id': 123}

        self.cg._tag_cg_build()

        koji_session.getTag.assert_called_once_with(self.cg.module.cg_build_koji_tag)
        koji_session.tagBuild.assert_called_once_with(123, "nginx-0-2.10e50d06")

        # tagBuild requires logging into a session in advance.
        koji_session.krb_login.assert_called_once()

    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    def test_tag_cg_build_fallback_to_default_tag(self, ClientSession):
        """ Test that the CG build is tagged to default tag. """
        koji_session = ClientSession.return_value
        koji_session.getUser.return_value = GET_USER_RV
        koji_session.getTag.side_effect = [{}, {'id': 123}]

        self.cg._tag_cg_build()

        assert koji_session.getTag.mock_calls == [
            call(self.cg.module.cg_build_koji_tag),
            call(conf.koji_cg_default_build_tag)]
        koji_session.tagBuild.assert_called_once_with(123, "nginx-0-2.10e50d06")

        # tagBuild requires logging into a session in advance.
        koji_session.krb_login.assert_called_once()

    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    def test_tag_cg_build_no_tag_set(self, ClientSession):
        """ Test that the CG build is not tagged when no tag set. """
        koji_session = ClientSession.return_value
        koji_session.getUser.return_value = GET_USER_RV
        koji_session.getTag.side_effect = [{}, {'id': 123}]

        self.cg.module.cg_build_koji_tag = None
        self.cg._tag_cg_build()

        koji_session.tagBuild.assert_not_called()
        # tagBuild requires logging into a session in advance.
        koji_session.krb_login.assert_called_once()

    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    def test_tag_cg_build_no_tag_available(self, ClientSession):
        """ Test that the CG build is not tagged when no tag available. """
        koji_session = ClientSession.return_value
        koji_session.getUser.return_value = GET_USER_RV
        koji_session.getTag.side_effect = [{}, {}]

        self.cg._tag_cg_build()

        koji_session.tagBuild.assert_not_called()
        # tagBuild requires logging into a session in advance.
        koji_session.krb_login.assert_called_once()

    @patch("module_build_service.builder.KojiContentGenerator.open", create=True)
    def test_get_arch_mmd_output(self, patched_open):
        patched_open.return_value = mock_open(
            read_data=self.cg.mmd).return_value
        ret = self.cg._get_arch_mmd_output("./fake-dir", "x86_64")
        assert ret == {
            'arch': 'x86_64',
            'buildroot_id': 1,
            'checksum': 'bf1615b15f6a0fee485abe94af6b56b6',
            'checksum_type': 'md5',
            'components': [],
            'extra': {'typeinfo': {'module': {}}},
            'filename': 'modulemd.x86_64.txt',
            'filesize': 1134,
            'type': 'file'
        }

    @patch("module_build_service.builder.KojiContentGenerator.open", create=True)
    def test_get_arch_mmd_output_components(self, patched_open):
        mmd = self.cg.module.mmd()
        rpm_artifacts = mmd.get_rpm_artifacts()
        rpm_artifacts.add("dhcp-libs-12:4.3.5-5.module_2118aef6.x86_64")
        mmd.set_rpm_artifacts(rpm_artifacts)
        mmd_data = mmd.dumps()

        patched_open.return_value = mock_open(
            read_data=mmd_data).return_value

        self.cg.rpms = [{
            "name": "dhcp",
            "version": "4.3.5",
            "release": "5.module_2118aef6",
            "arch": "x86_64",
            "epoch": "12",
            "payloadhash": "hash",
        }]

        self.cg.rpms_dict = {
            "dhcp-libs-12:4.3.5-5.module_2118aef6.x86_64": {
                "name": "dhcp",
                "version": "4.3.5",
                "release": "5.module_2118aef6",
                "arch": "x86_64",
                "epoch": "12",
                "payloadhash": "hash",
            }
        }

        ret = self.cg._get_arch_mmd_output("./fake-dir", "x86_64")
        assert ret == {
            'arch': 'x86_64',
            'buildroot_id': 1,
            'checksum': '1bcc38b6f19285b3656b84a0443f46d2',
            'checksum_type': 'md5',
            'components': [{u'arch': 'x86_64',
                            u'epoch': '12',
                            u'name': 'dhcp',
                            u'release': '5.module_2118aef6',
                            u'sigmd5': 'hash',
                            u'type': u'rpm',
                            u'version': '4.3.5'}],
            'extra': {'typeinfo': {'module': {}}},
            'filename': 'modulemd.x86_64.txt',
            'filesize': 315,
            'type': 'file'
        }

    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    def test_koji_rpms_in_tag(self, ClientSession):
        koji_session = ClientSession.return_value
        koji_session.getUser.return_value = GET_USER_RV
        koji_session.getTag.return_value = {"arches": "x86_64"}

        rpms = [
            {
                'id': 1,
                'arch': 'src',
                'epoch': None,
                'build_id': 875991,
                'name': 'module-build-macros',
                'release': '1.module_92011fe6',
                'version': '0.1'
            },
            {
                'id': 2,
                'arch': 'noarch',
                'epoch': None,
                'build_id': 875991,
                'name': 'module-build-macros',
                'release': '1.module_92011fe6',
                'version': '0.1'
            },
            {
                'id': 3,
                'arch': 'src',
                'epoch': 3,
                'build_id': 875636,
                'name': 'ed',
                'release': '2.module_bd6e0eb1',
                'version': '1.14.1'
            },
            {
                'id': 4,
                'arch': 'x86_64',
                'epoch': 3,
                'build_id': 875636,
                'name': 'ed',
                'release': '2.module_bd6e0eb1',
                'version': '1.14.1'
            },
        ]

        builds = [
            {
                'build_id': 875636,
                'epoch': 3,
                'name': 'ed',
                'release': '2.module_bd6e0eb1',
                'version': '1.14.1',
                'nvr': 'ed-2.module_bd6e0eb1-1.14.1',
            },
            {
                'build_id': 875991,
                'epoch': None,
                'name': 'module-build-macros',
                'release': '1.module_92011fe6',
                'version': '0.1',
                'nvr': 'module-build-macros-0.1-1.module_92011fe6',
            }
        ]

        koji_session.listTaggedRPMS.return_value = (rpms, builds)
        koji_session.multiCall.side_effect = [
            # getRPMHeaders response
            [[{'excludearch': ["x86_64"], 'exclusivearch': [], 'license': 'MIT'}],
             [{'excludearch': [], 'exclusivearch': ["x86_64"], 'license': 'GPL'}],
             [{'license': 'MIT'}],
             [{'license': 'GPL'}]]
        ]

        rpms = self.cg._koji_rpms_in_tag("tag")
        for rpm in rpms:
            # We want to mainly check the excludearch and exclusivearch code.
            if rpm["name"] == "module-build-macros":
                assert rpm["srpm_nevra"] == "module-build-macros-0:0.1-1.module_92011fe6.src"
                assert rpm["excludearch"] == ["x86_64"]
                assert rpm["license"] == "MIT"
            else:
                assert rpm["srpm_nevra"] == "ed-3:1.14.1-2.module_bd6e0eb1.src"
                assert rpm["exclusivearch"] == ["x86_64"]
                assert rpm["license"] == "GPL"

        # Listing tagged RPMs does not require to log into a session
        koji_session.krb_login.assert_not_called()

    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    def test_koji_rpms_in_tag_empty_tag(self, ClientSession):
        koji_session = ClientSession.return_value
        koji_session.getUser.return_value = GET_USER_RV
        koji_session.getTag.return_value = {"arches": "x86_64"}
        koji_session.listTaggedRPMS.return_value = ([], [])
        koji_session.multiCall.side_effect = [[], [], [], []]

        rpms = self.cg._koji_rpms_in_tag("tag")
        assert rpms == []
        koji_session.multiCall.assert_not_called()

    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    def test_koji_rpms_in_tag_empty_headers(self, ClientSession):
        koji_session = ClientSession.return_value
        koji_session.getUser.return_value = GET_USER_RV
        koji_session.getTag.return_value = {"arches": "x86_64"}

        rpms = [
            {
                'id': 1,
                'arch': 'src',
                'epoch': None,
                'build_id': 875991,
                'name': 'module-build-macros',
                'release': '1.module_92011fe6',
                'version': '0.1'
            },
            {
                'id': 2,
                'arch': 'noarch',
                'epoch': None,
                'build_id': 875991,
                'name': 'module-build-macros',
                'release': '1.module_92011fe6',
                'version': '0.1'
            },
        ]

        builds = [
            {
                'build_id': 875991,
                'epoch': None,
                'name': 'module-build-macros',
                'release': '1.module_92011fe6',
                'version': '0.1',
                'nvr': 'module-build-macros-0.1-1.module_92011fe6',
            }
        ]

        koji_session.listTaggedRPMS.return_value = (rpms, builds)

        koji_session.multiCall.side_effect = [
            # getRPMHeaders response
            [[{}], [{}]]
        ]

        with pytest.raises(RuntimeError) as cm:
            self.cg._koji_rpms_in_tag("tag")
        assert str(cm.value) == (
            "No RPM headers received from Koji for RPM module-build-macros")

        koji_session.multiCall.side_effect = [
            # getRPMHeaders response
            [[{"something": "x"}], [{}]]
        ]

        with pytest.raises(RuntimeError) as cm:
            self.cg._koji_rpms_in_tag("tag")
        assert str(cm.value) == (
            "No RPM 'license' header received from Koji for RPM module-build-macros")

    def _add_test_rpm(self, nevra, srpm_nevra, multilib=None,
                      koji_srpm_nevra=None, excludearch=None, exclusivearch=None,
                      license=None):
        """
        Helper method to add test RPM to ModuleBuild used by KojiContentGenerator
        and also to Koji tag used to generate the Content Generator build.

        :param str nevra: NEVRA of the RPM to add.
        :param str srpm_nevra: NEVRA of SRPM the added RPM is built from.
        :param list multilib: List of architecture for which the multilib should be turned on.
        :param str koji_srpm_nevra: If set, overrides the `srpm_nevra` in Koji tag. This is
            needed to test the case when the built "src" package has different name than
            the package in Koji. This is for example case of software collections where
            the name in `srpm_nevra` is "httpd" but `koji_srpm_nevra` name
            would be "httpd24-httpd".
        :param list excludearch: List of architectures this package is excluded from.
        :param list exclusivearch: List of architectures this package is exclusive for.
        :param str license: License of this RPM.
        """
        srpm_name = kobo.rpmlib.parse_nvra(srpm_nevra)["name"]

        parsed_nevra = kobo.rpmlib.parse_nvra(nevra)
        parsed_nevra["payloadhash"] = "hash"
        if koji_srpm_nevra:
            parsed_nevra["srpm_nevra"] = koji_srpm_nevra
            parsed_nevra["srpm_name"] = kobo.rpmlib.parse_nvra(koji_srpm_nevra)["name"]
        else:
            parsed_nevra["srpm_nevra"] = srpm_nevra
            parsed_nevra["srpm_name"] = srpm_name
        parsed_nevra["excludearch"] = excludearch or []
        parsed_nevra["exclusivearch"] = exclusivearch or []
        parsed_nevra["license"] = license or ""
        self.cg.rpms.append(parsed_nevra)
        self.cg.rpms_dict[nevra] = parsed_nevra

        mmd = self.cg.module.mmd()
        if srpm_name not in mmd.get_rpm_components().keys():
            component = Modulemd.ComponentRpm()
            component.set_name(srpm_name)
            component.set_rationale("foo")

            if multilib:
                multilib_set = Modulemd.SimpleSet()
                for arch in multilib:
                    multilib_set.add(arch)
                component.set_multilib(multilib_set)

            mmd.add_rpm_component(component)
            self.cg.module.modulemd = mmd.dumps()
            self.cg.modulemd = mmd.dumps()

    @pytest.mark.parametrize("devel", (False, True))
    def test_fill_in_rpms_list(self, devel):
        self._add_test_rpm("dhcp-12:4.3.5-5.module_2118aef6.src",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("dhcp-libs-12:4.3.5-5.module_2118aef6.x86_64",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("dhcp-libs-12:4.3.5-5.module_2118aef6.i686",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("dhcp-libs-12:4.3.5-5.module_2118aef6.s390x",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.x86_64",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.i686",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.s390x",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")

        self.cg.devel = devel
        mmd = self.cg.module.mmd()
        mmd = self.cg._fill_in_rpms_list(mmd, "x86_64")

        if not devel:
            # Only x86_64 packages should be filled in, because we requested x86_64 arch.
            assert set(mmd.get_rpm_artifacts().get()) == set([
                "dhcp-12:4.3.5-5.module_2118aef6.src",
                "dhcp-libs-12:4.3.5-5.module_2118aef6.x86_64",
                "perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                "perl-Tangerine-12:4.3.5-5.module_2118aef6.x86_64",
            ])
        else:
            # The i686 packages are filtered out in normal packages, because multilib
            # is not enabled for them - therefore we want to include them in -devel.
            assert set(mmd.get_rpm_artifacts().get()) == set([
                "dhcp-libs-12:4.3.5-5.module_2118aef6.i686",
                "perl-Tangerine-12:4.3.5-5.module_2118aef6.i686"])

    def test_fill_in_rpms_exclusivearch(self):
        self._add_test_rpm("dhcp-12:4.3.5-5.module_2118aef6.src",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("dhcp-libs-12:4.3.5-5.module_2118aef6.noarch",
                           "dhcp-12:4.3.5-5.module_2118aef6.src",
                           exclusivearch=["x86_64"])
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.noarch",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           exclusivearch=["ppc64le"])

        mmd = self.cg.module.mmd()
        mmd = self.cg._fill_in_rpms_list(mmd, "x86_64")

        # Only dhcp-libs should be filled in, because perl-Tangerine has different
        # exclusivearch.
        assert set(mmd.get_rpm_artifacts().get()) == set([
            "dhcp-12:4.3.5-5.module_2118aef6.src",
            "dhcp-libs-12:4.3.5-5.module_2118aef6.noarch",
        ])

    def test_fill_in_rpms_excludearch(self):
        self._add_test_rpm("dhcp-12:4.3.5-5.module_2118aef6.src",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("dhcp-libs-12:4.3.5-5.module_2118aef6.noarch",
                           "dhcp-12:4.3.5-5.module_2118aef6.src",
                           excludearch=["x86_64"])
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.noarch",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           excludearch=["ppc64le"])

        mmd = self.cg.module.mmd()
        mmd = self.cg._fill_in_rpms_list(mmd, "x86_64")

        # Only perl-Tangerine should be filled in, because dhcp-libs is excluded from x86_64.
        assert set(mmd.get_rpm_artifacts().get()) == set([
            "perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
            "perl-Tangerine-12:4.3.5-5.module_2118aef6.noarch",
        ])

    @pytest.mark.parametrize("devel", (False, True))
    def test_fill_in_rpms_rpm_whitelist(self, devel):
        self._add_test_rpm("python27-dhcp-12:4.3.5-5.module_2118aef6.src",
                           "dhcp-12:4.3.5-5.module_2118aef6.src",
                           koji_srpm_nevra="python27-dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("python27-dhcp-libs-12:4.3.5-5.module_2118aef6.x86_64",
                           "dhcp-12:4.3.5-5.module_2118aef6.src",
                           koji_srpm_nevra="python27-dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("python27-dhcp-libs-12:4.3.5-5.module_2118aef6.i686",
                           "dhcp-12:4.3.5-5.module_2118aef6.src",
                           koji_srpm_nevra="python27-dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("foo-perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           koji_srpm_nevra="foo-perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("foo-perl-Tangerine-12:4.3.5-5.module_2118aef6.x86_64",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           koji_srpm_nevra="foo-perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("foo-perl-Tangerine-12:4.3.5-5.module_2118aef6.i686",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           koji_srpm_nevra="foo-perl-Tangerine-12:4.3.5-5.module_2118aef6.src")

        self.cg.devel = devel
        mmd = self.cg.module.mmd()
        opts = mmd.get_buildopts()
        opts.set_rpm_whitelist(["python27-dhcp"])
        mmd.set_buildopts(opts)

        mmd = self.cg._fill_in_rpms_list(mmd, "x86_64")

        if not devel:
            # Only x86_64 dhcp-libs should be filled in, because only python27-dhcp is whitelisted
            # srpm name.
            assert set(mmd.get_rpm_artifacts().get()) == set([
                "python27-dhcp-12:4.3.5-5.module_2118aef6.src",
                "python27-dhcp-libs-12:4.3.5-5.module_2118aef6.x86_64",
            ])
        else:
            assert set(mmd.get_rpm_artifacts().get()) == set([
                "python27-dhcp-libs-12:4.3.5-5.module_2118aef6.i686",
                "foo-perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                "foo-perl-Tangerine-12:4.3.5-5.module_2118aef6.x86_64",
                "foo-perl-Tangerine-12:4.3.5-5.module_2118aef6.i686",
            ])

    @pytest.mark.parametrize("devel", (False, True))
    def test_fill_in_rpms_list_filters(self, devel):
        self._add_test_rpm("dhcp-12:4.3.5-5.module_2118aef6.src",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("dhcp-libs-12:4.3.5-5.module_2118aef6.x86_64",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("dhcp-libs-debuginfo-12:4.3.5-5.module_2118aef6.x86_64",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("dhcp-libs-debugsource-12:4.3.5-5.module_2118aef6.x86_64",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("dhcp-libs-12:4.3.5-5.module_2118aef6.i686",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("dhcp-libs-debuginfo-12:4.3.5-5.module_2118aef6.i686",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("dhcp-libs-debugsource-12:4.3.5-5.module_2118aef6.i686",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.x86_64",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-debuginfo-12:4.3.5-5.module_2118aef6.x86_64",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-debugsource-12:4.3.5-5.module_2118aef6.x86_64",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.i686",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-debuginfo-12:4.3.5-5.module_2118aef6.i686",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-debugsource-12:4.3.5-5.module_2118aef6.i686",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")

        self.cg.devel = devel
        mmd = self.cg.module.mmd()
        filter_list = Modulemd.SimpleSet()
        filter_list.add("dhcp-libs")
        mmd.set_rpm_filter(filter_list)

        mmd = self.cg._fill_in_rpms_list(mmd, "x86_64")

        if not devel:
            # Only x86_64 perl-Tangerine should be filled in, because dhcp-libs is filtered out.
            assert set(mmd.get_rpm_artifacts().get()) == set([
                "perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                "perl-Tangerine-12:4.3.5-5.module_2118aef6.x86_64",
                "perl-Tangerine-debuginfo-12:4.3.5-5.module_2118aef6.x86_64",
                "perl-Tangerine-debugsource-12:4.3.5-5.module_2118aef6.x86_64",
            ])
        else:
            assert set(mmd.get_rpm_artifacts().get()) == set([
                "dhcp-12:4.3.5-5.module_2118aef6.src",
                "dhcp-libs-12:4.3.5-5.module_2118aef6.x86_64",
                "dhcp-libs-debuginfo-12:4.3.5-5.module_2118aef6.x86_64",
                "dhcp-libs-debugsource-12:4.3.5-5.module_2118aef6.x86_64",
                "dhcp-libs-12:4.3.5-5.module_2118aef6.i686",
                "dhcp-libs-debuginfo-12:4.3.5-5.module_2118aef6.i686",
                "dhcp-libs-debugsource-12:4.3.5-5.module_2118aef6.i686",
                "perl-Tangerine-12:4.3.5-5.module_2118aef6.i686",
                "perl-Tangerine-debuginfo-12:4.3.5-5.module_2118aef6.i686",
                "perl-Tangerine-debugsource-12:4.3.5-5.module_2118aef6.i686",
            ])

    @pytest.mark.parametrize("devel", (False, True))
    def test_fill_in_rpms_list_multilib(self, devel):
        self._add_test_rpm("dhcp-libs-12:4.3.5-5.module_2118aef6.src",
                           "dhcp-libs-12:4.3.5-5.module_2118aef6.src",
                           multilib=["x86_64"])
        self._add_test_rpm("dhcp-libs-12:4.3.5-5.module_2118aef6.x86_64",
                           "dhcp-libs-12:4.3.5-5.module_2118aef6.src",
                           multilib=["x86_64"])
        self._add_test_rpm("dhcp-libs-12:4.3.5-5.module_2118aef6.i686",
                           "dhcp-libs-12:4.3.5-5.module_2118aef6.src",
                           multilib=["x86_64"])
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           multilib=["ppc64le"])
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.x86_64",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           multilib=["ppc64le"])
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.i686",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           multilib=["ppc64le"])

        self.cg.devel = devel
        mmd = self.cg.module.mmd()
        mmd = self.cg._fill_in_rpms_list(mmd, "x86_64")

        if not devel:
            # Only i686 package for dhcp-libs should be added, because perl-Tangerine does not have
            # multilib set.
            assert set(mmd.get_rpm_artifacts().get()) == set([
                "dhcp-libs-12:4.3.5-5.module_2118aef6.src",
                "dhcp-libs-12:4.3.5-5.module_2118aef6.x86_64",
                "dhcp-libs-12:4.3.5-5.module_2118aef6.i686",
                "perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                "perl-Tangerine-12:4.3.5-5.module_2118aef6.x86_64",
            ])
        else:
            assert set(mmd.get_rpm_artifacts().get()) == set([
                "perl-Tangerine-12:4.3.5-5.module_2118aef6.i686"])

    @pytest.mark.parametrize(
        "licenses, expected", (
            (["GPL", "MIT"], ["GPL", "MIT"]),
            (["GPL", ""], ["GPL"]),
            (["GPL", "GPL"], ["GPL"]),
        )
    )
    def test_fill_in_rpms_list_license(self, licenses, expected):
        self._add_test_rpm("dhcp-12:4.3.5-5.module_2118aef6.src",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("dhcp-libs-12:4.3.5-5.module_2118aef6.x86_64",
                           "dhcp-12:4.3.5-5.module_2118aef6.src",
                           license=licenses[0])
        self._add_test_rpm("dhcp-libs-12:4.3.5-5.module_2118aef6.i686",
                           "dhcp-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.x86_64",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src",
                           license=licenses[1])
        self._add_test_rpm("perl-Tangerine-12:4.3.5-5.module_2118aef6.i686",
                           "perl-Tangerine-12:4.3.5-5.module_2118aef6.src")

        mmd = self.cg.module.mmd()
        mmd = self.cg._fill_in_rpms_list(mmd, "x86_64")

        # Only x86_64 packages should be filled in, because we requested x86_64 arch.
        assert set(mmd.get_content_licenses().get()) == set(expected)

    @pytest.mark.parametrize("devel", (False, True))
    def test_fill_in_rpms_list_noarch_filtering_not_influenced_by_multilib(self, devel):
        # A build has ExcludeArch: i686 (because it only works on 64 bit arches).
        # A noarch package is built there, and this noarch packages should be
        # included in x86_64 repo.
        self._add_test_rpm("dhcp-libs-12:4.3.5-5.module_2118aef6.noarch",
                           "dhcp-12:4.3.5-5.module_2118aef6.src",
                           excludearch=["i686"])
        self._add_test_rpm("dhcp-12:4.3.5-5.module_2118aef6.src",
                           "dhcp-12:4.3.5-5.module_2118aef6.src",
                           excludearch=["i686"])

        self.cg.devel = devel
        mmd = self.cg.module.mmd()
        mmd = self.cg._fill_in_rpms_list(mmd, "x86_64")

        if not devel:
            # Only i686 package for dhcp-libs should be added, because perl-Tangerine does not have
            # multilib set. The "dhcp" SRPM should be also included.
            assert set(mmd.get_rpm_artifacts().get()) == set([
                "dhcp-libs-12:4.3.5-5.module_2118aef6.noarch",
                "dhcp-12:4.3.5-5.module_2118aef6.src"])
        else:
            assert set(mmd.get_rpm_artifacts().get()) == set([])

    def test_sanitize_mmd(self):
        mmd = self.cg.module.mmd()
        component = Modulemd.ComponentRpm()
        component.set_name("foo")
        component.set_rationale("foo")
        component.set_repository("http://private.tld/foo.git")
        component.set_cache("http://private.tld/cache")
        mmd.add_rpm_component(component)
        mmd.set_xmd(glib.dict_values({"mbs": {"buildrequires": []}}))
        mmd = self.cg._sanitize_mmd(mmd)

        for pkg in mmd.get_rpm_components().values():
            assert pkg.get_repository() is None
            assert pkg.get_cache() is None

        assert "mbs" not in mmd.get_xmd().keys()

    @patch('module_build_service.builder.KojiContentGenerator.SCM')
    def test_prepare_file_directory_modulemd_src(self, mocked_scm):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule_init.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')
        mmd = self.cg.module.mmd()
        mmd.set_xmd(glib.dict_values({"mbs": {
            "commit": "foo",
            "scmurl": "git://localhost/modules/foo.git#master"}}))
        self.cg.module.modulemd = mmd.dumps()
        file_dir = self.cg._prepare_file_directory()
        with open(path.join(file_dir, "modulemd.src.txt")) as mmd:
            assert len(mmd.read()) == 1337

    def test_finalize_mmd_devel(self):
        self.cg.devel = True
        mmd = self.cg.module.mmd()
        new_mmd = Modulemd.Module.new_from_string(self.cg._finalize_mmd("x86_64"))

        # Check that -devel suffix is set.
        assert new_mmd.get_name().endswith("-devel")

        # Check that -devel requires non-devel.
        for dep in new_mmd.get_dependencies():
            requires = []
            for name, streams in dep.get_requires().items():
                for stream in streams.get():
                    requires.append("%s:%s" % (name, stream))
            assert "%s:%s" % (mmd.get_name(), mmd.get_stream()) in requires



# flake8: noqa
import base64
import traceback

import koji
# Import krbV from here so we don't have to redo the whole try except that Koji does
from koji import krbV, PythonImportError, AuthError, AUTHTYPE_KERB


class ClientSession(koji.ClientSession):
    """The koji.ClientSession class with patches from upstream."""

    # This backport comes from https://pagure.io/koji/pull-request/1187
    def krb_login(self, principal=None, keytab=None, ccache=None, proxyuser=None, ctx=None):
        """Log in using Kerberos.  If principal is not None and keytab is
        not None, then get credentials for the given principal from the given keytab.
        If both are None, authenticate using existing local credentials (as obtained
        from kinit).  ccache is the absolute path to use for the credential cache. If
        not specified, the default ccache will be used.  If proxyuser is specified,
        log in the given user instead of the user associated with the Kerberos
        principal.  The principal must be in the "ProxyPrincipals" list on
        the server side.  ctx is the Kerberos context to use, and should be unique
        per thread.  If ctx is not specified, the default context is used."""

        try:
            # Silently try GSSAPI first
            if self.gssapi_login(principal, keytab, ccache, proxyuser=proxyuser):
                return True
        except Exception as e:
            if krbV:
                e_str = ''.join(traceback.format_exception_only(type(e), e))
                self.logger.debug('gssapi auth failed: %s', e_str)
                pass
            else:
                raise

        if not krbV:
            raise PythonImportError(
                "Please install python-krbV to use kerberos."
            )

        if not ctx:
            ctx = krbV.default_context()

        if ccache != None:
            ccache = krbV.CCache(name=ccache, context=ctx)
        else:
            ccache = ctx.default_ccache()

        if principal != None:
            if keytab != None:
                cprinc = krbV.Principal(name=principal, context=ctx)
                keytab = krbV.Keytab(name=keytab, context=ctx)
                ccache.init(cprinc)
                ccache.init_creds_keytab(principal=cprinc, keytab=keytab)
            else:
                raise AuthError('cannot specify a principal without a keytab')
        else:
            # We're trying to log ourself in.  Connect using existing credentials.
            cprinc = ccache.principal()

        self.logger.debug('Authenticating as: %s', cprinc.name)
        sprinc = krbV.Principal(name=self._serverPrincipal(cprinc), context=ctx)

        ac = krbV.AuthContext(context=ctx)
        ac.flags = krbV.KRB5_AUTH_CONTEXT_DO_SEQUENCE | krbV.KRB5_AUTH_CONTEXT_DO_TIME
        ac.rcache = ctx.default_rcache()

        # create and encode the authentication request
        (ac, req) = ctx.mk_req(server=sprinc, client=cprinc,
                               auth_context=ac, ccache=ccache,
                               options=krbV.AP_OPTS_MUTUAL_REQUIRED)
        req_enc = base64.encodestring(req)

        # ask the server to authenticate us
        (rep_enc, sinfo_enc, addrinfo) = self.callMethod('krbLogin', req_enc, proxyuser)
        # Set the addrinfo we received from the server
        # (necessary before calling rd_priv())
        # addrinfo is in (serveraddr, serverport, clientaddr, clientport)
        # format, so swap the pairs because clientaddr is now the local addr
        ac.addrs = tuple((addrinfo[2], addrinfo[3], addrinfo[0], addrinfo[1]))

        # decode and read the reply from the server
        rep = base64.decodestring(rep_enc)
        ctx.rd_rep(rep, auth_context=ac)

        # decode and decrypt the login info
        sinfo_priv = base64.decodestring(sinfo_enc)
        sinfo_str = ac.rd_priv(sinfo_priv)
        sinfo = dict(zip(['session-id', 'session-key'], sinfo_str.split()))

        if not sinfo:
            self.logger.warn('No session info received')
            return False
        self.setSession(sinfo)

        self.authtype = AUTHTYPE_KERB
        return True



# -*- coding: utf-8 -*-


# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

"""
Logging functions.

At the beginning of the MBS flow, init_logging(conf) must be called.

After that, logging from any module is possible using Python's "logging"
module as showed at
<https://docs.python.org/3/howto/logging.html#logging-basic-tutorial>.

Examples:

import logging

logging.debug("Phasers are set to stun.")
logging.info("%s tried to build something", username)
logging.warning("%s failed to build", task_id)

"""

import os
import logging

levels = {}
levels["debug"] = logging.DEBUG
levels["error"] = logging.ERROR
levels["warning"] = logging.WARNING
levels["info"] = logging.INFO

level_flags = {}
level_flags["debug"] = levels["debug"]
level_flags["verbose"] = levels["info"]
level_flags["quiet"] = levels["error"]


log_format = '%(asctime)s - %(threadName)s - %(name)s - %(levelname)s - %(message)s'


class ModuleBuildFileHandler(logging.FileHandler):
    """
    FileHandler subclass which handles only messages generated during
    particular module build with `build_id` set in its constructor.
    """
    def __init__(self, build_id, filename, mode='a', encoding=None, delay=0):
        logging.FileHandler.__init__(self, filename, mode, encoding, delay)
        self.build_id = build_id

    def emit(self, record):
        # Imported here because of importing cycle between __init__.py,
        # scheduler and models.
        from module_build_service.scheduler.consumer import MBSConsumer

        # Check the currently handled module build and emit the message only
        # if it's associated with current module.
        build_id = MBSConsumer.current_module_build_id
        if not build_id or build_id != self.build_id:
            return

        logging.FileHandler.emit(self, record)


class ModuleBuildLogs(object):
    """
    Manages ModuleBuildFileHandler logging handlers.
    """
    def __init__(self, build_logs_dir, build_logs_name_format, level=logging.INFO):
        """
        Creates new ModuleBuildLogs instance. Module build logs are stored
        to `build_logs_dir` directory.
        """
        self.handlers = {}
        self.build_logs_dir = build_logs_dir
        self.build_logs_name_format = build_logs_name_format
        self.level = level

    def path(self, build):
        """
        Returns the full path to build log of module with id `build_id`.
        """
        path = os.path.join(self.build_logs_dir, self.name(build))
        return path

    def name(self, build):
        """
        Returns the filename for a module build
        """
        name = self.build_logs_name_format.format(**build.json())
        return name

    def start(self, build):
        """
        Starts logging build log for module with `build_id` id.
        """
        if not self.build_logs_dir:
            return

        if build.id in self.handlers:
            return

        # Create and add ModuleBuildFileHandler.
        handler = ModuleBuildFileHandler(build.id, self.path(build))
        handler.setLevel(self.level)
        handler.setFormatter(logging.Formatter(log_format, None))
        log = logging.getLogger()
        log.setLevel(self.level)
        log.addHandler(handler)

        self.handlers[build.id] = handler

    def stop(self, build):
        """
        Stops logging build log for module with `build_id` id. It does *not*
        remove the build log from fs.
        """
        if build.id not in self.handlers:
            return

        handler = self.handlers[build.id]
        handler.flush()
        handler.close()

        # Remove the log handler.
        log = logging.getLogger()
        log.removeHandler(handler)
        del self.handlers[build.id]


def str_to_log_level(level):
    """
    Returns internal representation of logging level defined
    by the string `level`.

    Available levels are: debug, info, warning, error
    """
    if level not in levels:
        return logging.NOTSET

    return levels[level]


def supported_log_backends():
    return ("console", "file")


def init_logging(conf):
    """
    Initializes logging according to configuration file.
    """
    log_backend = conf.log_backend

    if not log_backend or len(log_backend) == 0 or log_backend == "console":
        logging.basicConfig(level=conf.log_level, format=log_format)
        log = logging.getLogger()
        log.setLevel(conf.log_level)
    else:
        logging.basicConfig(filename=conf.log_file, level=conf.log_level,
                            format=log_format)
        log = logging.getLogger()



import os
import shutil
import subprocess
import munch
import errno
import logging
from multiprocessing.dummy import Pool as ThreadPool

import requests

from module_build_service import log


logging.basicConfig(level=logging.DEBUG)


def find_srpm(cod):
    for f in os.listdir(cod):
        if f.endswith(".src.rpm"):
            return os.path.join(cod, f)


def execute_cmd(args, stdout=None, stderr=None, cwd=None):
    """
    Executes command defined by `args`. If `stdout` or `stderr` is set to
    Python file object, the stderr/stdout output is redirecter to that file.
    If `cwd` is set, current working directory is set accordingly for the
    executed command.

    :param args: list defining the command to execute.
    :param stdout: Python file object to redirect the stdout to.
    :param stderr: Python file object to redirect the stderr to.
    :param cwd: string defining the current working directory for command.
    :raises RuntimeError: Raised when command exits with non-zero exit code.
    """
    out_log_msg = ""
    if stdout and hasattr(stdout, "name"):
        out_log_msg += ", stdout log: %s" % stdout.name
    if stderr and hasattr(stderr, "name"):
        out_log_msg += ", stderr log: %s" % stderr.name

    log.info("Executing command: %s%s" % (args, out_log_msg))
    proc = subprocess.Popen(args, stdout=stdout, stderr=stderr, cwd=cwd)
    out, err = proc.communicate()

    if proc.returncode != 0:
        err_msg = "Command '%s' returned non-zero value %d%s" % (args, proc.returncode, out_log_msg)
        raise RuntimeError(err_msg)
    return out, err


def get_koji_config(mbs_config):
    """
    Get the Koji config needed for MBS
    :param mbs_config: an MBS config object
    :return: a Munch object of the Koji config
    """
    # Placed here to avoid py2/py3 conflicts...
    import koji

    koji_config = munch.Munch(koji.read_config(
        profile_name=mbs_config.koji_profile,
        user_config=mbs_config.koji_config,
    ))
    # Timeout after 10 minutes.  The default is 12 hours.
    koji_config["timeout"] = 60 * 10
    return koji_config


def create_local_repo_from_koji_tag(config, tag, repo_dir, archs=None):
    """
    Downloads the packages build for one of `archs` (defaults to ['x86_64',
    'noarch']) in Koji tag `tag` to `repo_dir` and creates repository in that
    directory. Needs config.koji_profile and config.koji_config to be set.
    """

    # Placed here to avoid py2/py3 conflicts...
    import koji

    if not archs:
        archs = ["x86_64", "noarch"]

    # Load koji config and create Koji session.
    koji_config = get_koji_config(config)
    address = koji_config.server
    log.info("Connecting to koji %r" % address)
    session = koji.ClientSession(address, opts=koji_config)

    # Get the list of all RPMs and builds in a tag.
    try:
        rpms, builds = session.listTaggedRPMS(tag, latest=True)
    except koji.GenericError:
        log.exception("Failed to list rpms in tag %r" % tag)

    # Reformat builds so they are dict with build_id as a key.
    builds = {build['build_id']: build for build in builds}

    # Prepare pathinfo we will use to generate the URL.
    pathinfo = koji.PathInfo(topdir=session.opts["topurl"])

    # When True, we want to run the createrepo_c.
    repo_changed = False

    # Prepare the list of URLs to download
    download_args = []
    for rpm in rpms:
        build_info = builds[rpm['build_id']]

        # We do not download debuginfo packages or packages built for archs
        # we are not interested in.
        if koji.is_debuginfo(rpm['name']) or not rpm['arch'] in archs:
            continue

        fname = pathinfo.rpm(rpm)
        relpath = os.path.basename(fname)
        local_fn = os.path.join(repo_dir, relpath)
        # Download only when the RPM is not downloaded or the size does not match.
        if not os.path.exists(local_fn) or os.path.getsize(local_fn) != rpm['size']:
            if os.path.exists(local_fn):
                os.remove(local_fn)
            repo_changed = True
            url = pathinfo.build(build_info) + '/' + fname
            download_args.append((url, local_fn))

    log.info(
        "Downloading %d packages from Koji tag %s to %s" % (len(download_args), tag, repo_dir))

    # Create the output directory
    try:
        os.makedirs(repo_dir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise

    def _download_file(url_and_dest):
        """
        Download a file in a memory efficient manner
        :param url_and_dest: a tuple containing the URL and the destination to download to
        :return: None
        """
        log.info("Downloading {0}...".format(url_and_dest[0]))
        if len(url_and_dest) != 2:
            raise ValueError("url_and_dest must have two values")

        rv = requests.get(url_and_dest[0], stream=True, timeout=60)
        with open(url_and_dest[1], "wb") as f:
            for chunk in rv.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)

    # Download the RPMs four at a time.
    pool = ThreadPool(4)
    try:
        pool.map(_download_file, download_args)
    finally:
        pool.close()

    # If we downloaded something, run the createrepo_c.
    if repo_changed:
        repodata_path = os.path.join(repo_dir, "repodata")
        if os.path.exists(repodata_path):
            shutil.rmtree(repodata_path)

        log.info("Creating local repository in %s" % repo_dir)
        execute_cmd(['/usr/bin/createrepo_c', repo_dir])



# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>

import os
import shutil
import tempfile

import pytest

import module_build_service.scm
from module_build_service.errors import ValidationError, UnprocessableEntity

base_dir = os.path.join(os.path.dirname(__file__), 'scm_data')
repo_url = 'file://' + base_dir + '/testrepo'


class TestSCMModule:

    def setup_method(self, test_method):
        self.tempdir = tempfile.mkdtemp()
        self.repodir = self.tempdir + '/testrepo'

    def teardown_method(self, test_method):
        if os.path.exists(self.tempdir):
            shutil.rmtree(self.tempdir)

    def test_simple_local_checkout(self):
        """ See if we can clone a local git repo. """
        scm = module_build_service.scm.SCM(repo_url)
        scm.checkout(self.tempdir)
        files = os.listdir(self.repodir)
        assert 'foo' in files, "foo not in %r" % files

    def test_local_get_latest_is_sane(self):
        """ See that a hash is returned by scm.get_latest. """
        scm = module_build_service.scm.SCM(repo_url)
        latest = scm.get_latest('master')
        target = '5481faa232d66589e660cc301179867fb00842c9'
        assert latest == target, "%r != %r" % (latest, target)

    def test_local_get_latest_unclean_input(self):
        """ Ensure that shell characters aren't handled poorly.

        https://pagure.io/fm-orchestrator/issue/329
        """
        scm = module_build_service.scm.SCM(repo_url)
        assert scm.scheme == 'git', scm.scheme
        fname = tempfile.mktemp(suffix='mbs-scm-test')
        try:
            scm.get_latest('master; touch %s' % fname)
        except UnprocessableEntity:
            assert not os.path.exists(fname), "%r exists!  Vulnerable." % fname

    def test_local_extract_name(self):
        scm = module_build_service.scm.SCM(repo_url)
        target = 'testrepo'
        assert scm.name == target, '%r != %r' % (scm.name, target)

    def test_local_extract_name_trailing_slash(self):
        scm = module_build_service.scm.SCM(repo_url + '/')
        target = 'testrepo'
        assert scm.name == target, '%r != %r' % (scm.name, target)

    def test_verify(self):
        scm = module_build_service.scm.SCM(repo_url)
        scm.checkout(self.tempdir)
        scm.verify()

    def test_verify_unknown_branch(self):
        with pytest.raises(UnprocessableEntity):
            module_build_service.scm.SCM(repo_url, "unknown")

    def test_verify_commit_in_branch(self):
        target = '7035bd33614972ac66559ac1fdd019ff6027ad21'
        scm = module_build_service.scm.SCM(repo_url + "?#" + target, "dev")
        scm.checkout(self.tempdir)
        scm.verify()

    def test_verify_commit_not_in_branch(self):
        target = '7035bd33614972ac66559ac1fdd019ff6027ad21'
        scm = module_build_service.scm.SCM(repo_url + "?#" + target, "master")
        scm.checkout(self.tempdir)
        with pytest.raises(ValidationError):
            scm.verify()

    def test_verify_unknown_hash(self):
        target = '7035bd33614972ac66559ac1fdd019ff6027ad22'
        scm = module_build_service.scm.SCM(repo_url + "?#" + target, "master")
        with pytest.raises(UnprocessableEntity):
            scm.checkout(self.tempdir)

    def test_get_module_yaml(self):
        scm = module_build_service.scm.SCM(repo_url)
        scm.checkout(self.tempdir)
        scm.verify()
        with pytest.raises(UnprocessableEntity):
            scm.get_module_yaml()

    def test_get_latest_incorrect_component_branch(self):
        scm = module_build_service.scm.SCM(repo_url)
        with pytest.raises(UnprocessableEntity):
            scm.get_latest('foobar')

    def test_get_latest_component_branch(self):
        ref = "5481faa232d66589e660cc301179867fb00842c9"
        branch = "master"
        scm = module_build_service.scm.SCM(repo_url)
        commit = scm.get_latest(branch)
        assert commit == ref

    def test_get_latest_component_ref(self):
        ref = "5481faa232d66589e660cc301179867fb00842c9"
        scm = module_build_service.scm.SCM(repo_url)
        commit = scm.get_latest(ref)
        assert commit == ref

    def test_get_latest_incorrect_component_ref(self):
        scm = module_build_service.scm.SCM(repo_url)
        with pytest.raises(UnprocessableEntity):
            scm.get_latest('15481faa232d66589e660cc301179867fb00842c9')



# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os

import pytest

import module_build_service.utils
from module_build_service import glib
from module_build_service.errors import StreamAmbigous
from tests import db, clean_database, make_module, init_data, base_dir


class TestUtilsModuleStreamExpansion:

    def setup_method(self, test_method):
        clean_database(False)

    def teardown_method(self, test_method):
        clean_database()

    def _get_mmds_required_by_module_recursively(self, module_build):
        """
        Convenience wrapper around get_mmds_required_by_module_recursively
        returning the list with nsvc strings of modules returned by this the wrapped
        method.
        """
        mmd = module_build.mmd()
        module_build_service.utils.expand_mse_streams(db.session, mmd)
        modules = module_build_service.utils.get_mmds_required_by_module_recursively(mmd)
        nsvcs = [":".join([m.get_name(), m.get_stream(), str(m.get_version()), m.get_context()])
                 for m in modules]
        return nsvcs

    def _generate_default_modules(self):
        """
        Generates gtk:1, gtk:2, foo:1 and foo:2 modules requiring the
        platform:f28 and platform:f29 modules.
        """
        platform_f28 = make_module("platform:f28:0:c10", {}, {})
        platform_f29 = make_module("platform:f29:0:c11", {}, {})
        make_module("gtk:1:0:c2", {"platform": ["f28"]}, {}, platform_f28)
        make_module("gtk:1:0:c3", {"platform": ["f29"]}, {}, platform_f29)
        make_module("gtk:2:0:c4", {"platform": ["f28"]}, {}, platform_f28)
        make_module("gtk:2:0:c5", {"platform": ["f29"]}, {}, platform_f29)
        make_module("foo:1:0:c2", {"platform": ["f28"]}, {}, platform_f28)
        make_module("foo:1:0:c3", {"platform": ["f29"]}, {}, platform_f29)
        make_module("foo:2:0:c4", {"platform": ["f28"]}, {}, platform_f28)
        make_module("foo:2:0:c5", {"platform": ["f29"]}, {}, platform_f29)
        make_module("app:1:0:c6", {"platform": ["f29"]}, {}, platform_f29)

    def test_generate_expanded_mmds_context(self):
        self._generate_default_modules()
        module_build = make_module(
            "app:1:0:c1", {"gtk": ["1", "2"]}, {"platform": ["f28"], "gtk": ["1", "2"]})
        mmds = module_build_service.utils.generate_expanded_mmds(
            db.session, module_build.mmd())
        contexts = set([mmd.get_context() for mmd in mmds])
        assert set(['e1e005fb', 'ce132a1e']) == contexts

    @pytest.mark.parametrize(
        'requires,build_requires,stream_ambigous,expected_xmd,expected_buildrequires', [
            ({"gtk": ["1", "2"]},
             {"platform": ["f28"], "gtk": ["1", "2"]}, True,
             set([
                 frozenset(['platform:f28:0:c10', 'gtk:2:0:c4']),
                 frozenset(['platform:f28:0:c10', 'gtk:1:0:c2'])
             ]),
             set([
                 frozenset(['gtk:1', 'platform:f28']),
                 frozenset(['gtk:2', 'platform:f28']),
             ])),

            ({"foo": ["1"]},
             {"platform": ["f28"], "foo": ["1"], "gtk": ["1", "2"]}, True,
             set([
                 frozenset(['foo:1:0:c2', 'gtk:1:0:c2', 'platform:f28:0:c10']),
                 frozenset(['foo:1:0:c2', 'gtk:2:0:c4', 'platform:f28:0:c10'])
             ]),
             set([
                 frozenset(['foo:1', 'gtk:1', 'platform:f28']),
                 frozenset(['foo:1', 'gtk:2', 'platform:f28'])
             ])),

            ({"gtk": ["1"], "foo": ["1"]},
             {"platform": ["f28"], "gtk": ["1"], "foo": ["1"]}, False,
             set([
                 frozenset(['foo:1:0:c2', 'gtk:1:0:c2', 'platform:f28:0:c10'])
             ]),
             set([
                 frozenset(['foo:1', 'gtk:1', 'platform:f28'])
             ])),

            ({"gtk": ["1"], "foo": ["1"]},
             {"gtk": ["1"], "foo": ["1"], "platform": ["f28"]}, False,
             set([
                 frozenset(['foo:1:0:c2', 'gtk:1:0:c2', 'platform:f28:0:c10'])
             ]),
             set([
                 frozenset(['foo:1', 'gtk:1', 'platform:f28'])
             ])),

            ({"gtk": ["-2"], "foo": ["-2"]},
             {"platform": ["f28"], "gtk": ["-2"], "foo": ["-2"]}, True,
             set([
                 frozenset(['foo:1:0:c2', 'gtk:1:0:c2', 'platform:f28:0:c10'])
             ]),
             set([
                 frozenset(['foo:1', 'gtk:1', 'platform:f28'])
             ])),

            ({"gtk": ["1"], "foo": ["1"]},
             {"platform": ["f28"], "gtk": ["-1", "1"], "foo": ["-2", "1"]}, False,
             set([
                 frozenset(['foo:1:0:c2', 'gtk:1:0:c2', 'platform:f28:0:c10'])
             ]),
             set([
                 frozenset(['foo:1', 'gtk:1', 'platform:f28'])
             ])),

            ({"gtk": ["1"], "foo": ["1"]},
             {"platform": ["f28"], "gtk": ["1"]}, False,
             set([
                 frozenset(['gtk:1:0:c2', 'platform:f28:0:c10'])
             ]),
             set([
                 frozenset(['gtk:1', 'platform:f28'])
             ])),

            ({"gtk": []}, {"platform": ["f28"], "gtk": ["1"]}, True,
             set([
                 frozenset(['gtk:1:0:c2', 'platform:f28:0:c10'])
             ]),
             set([
                 frozenset(['gtk:1', 'platform:f28'])
             ])),

            ({}, {"platform": ["f29"], "app": ["1"]}, False,
             set([
                 frozenset(['app:1:0:c6', 'platform:f29:0:c11'])
             ]),
             set([
                 frozenset(['app:1', 'platform:f29'])
             ])),
        ])
    def test_generate_expanded_mmds_buildrequires(
            self, requires, build_requires, stream_ambigous, expected_xmd,
            expected_buildrequires):
        self._generate_default_modules()
        module_build = make_module("app:1:0:c1", requires, build_requires)

        # Check that generate_expanded_mmds raises an exception if stream is ambigous
        # and also that it does not raise an exception otherwise.
        if stream_ambigous:
            with pytest.raises(StreamAmbigous):
                module_build_service.utils.generate_expanded_mmds(
                    db.session, module_build.mmd(), raise_if_stream_ambigous=True)
        else:
            module_build_service.utils.generate_expanded_mmds(
                db.session, module_build.mmd(), raise_if_stream_ambigous=True)

        # Check that if stream is ambigous and we define the stream, it does not raise
        # an exception.
        if stream_ambigous:
            default_streams = {}
            for ns in list(expected_buildrequires)[0]:
                name, stream = ns.split(":")
                default_streams[name] = stream
            module_build_service.utils.generate_expanded_mmds(
                db.session, module_build.mmd(), raise_if_stream_ambigous=True,
                default_streams=default_streams)

        mmds = module_build_service.utils.generate_expanded_mmds(
            db.session, module_build.mmd())

        buildrequires_per_mmd_xmd = set()
        buildrequires_per_mmd_buildrequires = set()
        for mmd in mmds:
            xmd = glib.from_variant_dict(mmd.get_xmd())
            br_nsvcs = []
            for name, detail in xmd['mbs']['buildrequires'].items():
                br_nsvcs.append(":".join([
                    name, detail["stream"], detail["version"], detail["context"]]))
            buildrequires_per_mmd_xmd.add(frozenset(br_nsvcs))

            assert len(mmd.get_dependencies()) == 1

            buildrequires = set()
            dep = mmd.get_dependencies()[0]
            for req_name, req_streams in dep.get_buildrequires().items():
                for req_stream in req_streams.get():
                    buildrequires.add(":".join([req_name, req_stream]))
            buildrequires_per_mmd_buildrequires.add(frozenset(buildrequires))

        assert buildrequires_per_mmd_xmd == expected_xmd
        assert buildrequires_per_mmd_buildrequires == expected_buildrequires

    @pytest.mark.parametrize('requires,build_requires,expected', [
        ({"gtk": ["1", "2"]}, {"platform": [], "gtk": ["1", "2"]},
         set([
             frozenset(['gtk:1']),
             frozenset(['gtk:2']),
         ])),

        ({"gtk": ["1", "2"]}, {"platform": [], "gtk": ["1"]},
         set([
             frozenset(['gtk:1', 'gtk:2']),
         ])),

        ({"gtk": ["1"], "foo": ["1"]},
         {"platform": [], "gtk": ["1"], "foo": ["1"]},
         set([
             frozenset(['foo:1', 'gtk:1']),
         ])),

        ({"gtk": ["-2"], "foo": ["-2"]},
         {"platform": [], "gtk": ["-2"], "foo": ["-2"]},
         set([
             frozenset(['foo:1', 'gtk:1']),
         ])),

        ({"gtk": ["-1", "1"], "foo": ["-2", "1"]},
         {"platform": [], "gtk": ["-1", "1"], "foo": ["-2", "1"]},
         set([
             frozenset(['foo:1', 'gtk:1']),
         ])),

        ({"gtk": [], "foo": []}, {"platform": [], "gtk": ["1"], "foo": ["1"]},
         set([
             frozenset([]),
         ])),

    ])
    def test_generate_expanded_mmds_requires(self, requires, build_requires, expected):
        self._generate_default_modules()
        module_build = make_module("app:1:0:c1", requires, build_requires)
        mmds = module_build_service.utils.generate_expanded_mmds(
            db.session, module_build.mmd())

        requires_per_mmd = set()
        for mmd in mmds:
            assert len(mmd.get_dependencies()) == 1
            mmd_requires = set()
            dep = mmd.get_dependencies()[0]
            for req_name, req_streams in dep.get_requires().items():
                for req_stream in req_streams.get():
                    mmd_requires.add(":".join([req_name, req_stream]))
            requires_per_mmd.add(frozenset(mmd_requires))

        assert requires_per_mmd == expected

    @pytest.mark.parametrize('requires,build_requires,expected', [
        ({}, {"platform": [], "gtk": ["1", "2"]},
         ['platform:f29:0:c11', 'gtk:2:0:c4', 'gtk:2:0:c5',
          'platform:f28:0:c10', 'gtk:1:0:c2', 'gtk:1:0:c3']),

        ({}, {"platform": [], "gtk": ["1"], "foo": ["1"]},
         ['platform:f28:0:c10', 'gtk:1:0:c2', 'gtk:1:0:c3',
          'foo:1:0:c2', 'foo:1:0:c3', 'platform:f29:0:c11']),

        ({}, {"gtk": ["1"], "foo": ["1"], "platform": ["f28"]},
         ['platform:f28:0:c10', 'gtk:1:0:c2',
          'foo:1:0:c2']),

        ([{}, {}], [{"platform": [], "gtk": ["1"], "foo": ["1"]},
                    {"platform": [], "gtk": ["2"], "foo": ["2"]}],
         ['foo:1:0:c2', 'foo:1:0:c3', 'foo:2:0:c4', 'foo:2:0:c5',
          'platform:f28:0:c10', 'platform:f29:0:c11', 'gtk:1:0:c2',
          'gtk:1:0:c3', 'gtk:2:0:c4', 'gtk:2:0:c5']),

        ({}, {"platform": [], "gtk": ["-2"], "foo": ["-2"]},
         ['foo:1:0:c2', 'foo:1:0:c3', 'platform:f29:0:c11',
          'platform:f28:0:c10', 'gtk:1:0:c2', 'gtk:1:0:c3']),

        ({}, {"platform": [], "gtk": ["-1", "1"], "foo": ["-2", "1"]},
         ['foo:1:0:c2', 'foo:1:0:c3', 'platform:f29:0:c11',
          'platform:f28:0:c10', 'gtk:1:0:c2', 'gtk:1:0:c3']),
    ])
    def test_get_required_modules_simple(self, requires, build_requires, expected):
        module_build = make_module("app:1:0:c1", requires, build_requires)
        self._generate_default_modules()
        nsvcs = self._get_mmds_required_by_module_recursively(module_build)
        assert set(nsvcs) == set(expected)

    def _generate_default_modules_recursion(self):
        """
        Generates the gtk:1 module requiring foo:1 module requiring bar:1
        and lorem:1 modules which require base:f29 module requiring
        platform:f29 module :).
        """
        base_module = make_module("platform:f29:0:c11", {}, {})
        make_module("gtk:1:0:c2", {"foo": ["unknown"]}, {}, base_module)
        make_module("gtk:1:1:c2", {"foo": ["1"]}, {}, base_module)
        make_module("foo:1:0:c2", {"bar": ["unknown"]}, {}, base_module)
        make_module("foo:1:1:c2", {"bar": ["1"], "lorem": ["1"]}, {}, base_module)
        make_module("bar:1:0:c2", {"base": ["unknown"]}, {}, base_module)
        make_module("bar:1:1:c2", {"base": ["f29"]}, {}, base_module)
        make_module("lorem:1:0:c2", {"base": ["unknown"]}, {}, base_module)
        make_module("lorem:1:1:c2", {"base": ["f29"]}, {}, base_module)
        make_module("base:f29:0:c3", {"platform": ["f29"]}, {}, base_module)

    @pytest.mark.parametrize('requires,build_requires,expected', [
        ({}, {"platform": [], "gtk": ["1"]},
         ['foo:1:1:c2', 'base:f29:0:c3', 'platform:f29:0:c11',
          'bar:1:1:c2', 'gtk:1:1:c2', 'lorem:1:1:c2']),

        ({}, {"platform": [], "foo": ["1"]},
         ['foo:1:1:c2', 'base:f29:0:c3', 'platform:f29:0:c11',
          'bar:1:1:c2', 'lorem:1:1:c2']),
    ])
    def test_get_required_modules_recursion(self, requires, build_requires, expected):
        module_build = make_module("app:1:0:c1", requires, build_requires)
        self._generate_default_modules_recursion()
        nsvcs = self._get_mmds_required_by_module_recursively(module_build)
        assert set(nsvcs) == set(expected)

    def _generate_default_modules_modules_multiple_stream_versions(self):
        """
        Generates the gtk:1 module requiring foo:1 module requiring bar:1
        and lorem:1 modules which require base:f29 module requiring
        platform:f29 module :).
        """
        f290000 = make_module("platform:f29.0.0:0:c11", {}, {})
        f290100 = make_module("platform:f29.1.0:0:c11", {}, {})
        f290200 = make_module("platform:f29.2.0:0:c11", {}, {})
        make_module("gtk:1:0:c2", {"platform": ["f29"]}, {}, f290000)
        make_module("gtk:1:1:c2", {"platform": ["f29"]}, {}, f290100)
        make_module("gtk:1:2:c2", {"platform": ["f29"]}, {}, f290100)
        make_module("gtk:1:3:c2", {"platform": ["f29"]}, {}, f290200)

    @pytest.mark.parametrize('requires,build_requires,expected', [
        ({}, {"platform": ["f29.1.0"], "gtk": ["1"]},
         ['platform:f29.0.0:0:c11', 'gtk:1:0:c2', 'gtk:1:2:c2', 'platform:f29.1.0:0:c11']),
    ])
    def test_get_required_modules_stream_versions(self, requires, build_requires, expected):
        module_build = make_module("app:1:0:c1", requires, build_requires)
        self._generate_default_modules_modules_multiple_stream_versions()
        nsvcs = self._get_mmds_required_by_module_recursively(module_build)
        assert set(nsvcs) == set(expected)

    def test__get_base_module_mmds(self):
        """Ensure the correct results are returned without duplicates."""
        init_data(data_size=1, multiple_stream_versions=True)
        mmd = module_build_service.utils.load_mmd(
            os.path.join(base_dir, 'staged_data', 'testmodule_v2.yaml'), True)
        deps = mmd.get_dependencies()
        brs = deps[0].get_buildrequires()
        brs['platform'].set(['platform:f29.1.0', 'platform:f29.2.0'])
        deps[0].set_buildrequires(brs)
        mmd.set_dependencies(deps)

        mmds = module_build_service.utils.mse._get_base_module_mmds(mmd)
        expected = set(['platform:f29.0.0', 'platform:f29.1.0', 'platform:f29.2.0'])
        # Verify no duplicates were returned before doing set operations
        assert len(mmds) == len(expected)
        # Verify the expected ones were returned
        actual = set()
        for mmd_ in mmds:
            actual.add('{}:{}'.format(mmd_.get_name(), mmd_.get_stream()))
        assert actual == expected



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>

"""
Allows MBS to run behind reverse proxy and to ensure redirects work with https.

WSGI Middleware!!

Source: http://flask.pocoo.org/snippets/35/ by Peter Hansen
"""


class ReverseProxy(object):
    '''Wrap the application in this middleware and configure the
    front-end server to add these headers, to let you quietly bind
    this to a URL other than / and to an HTTP scheme that is
    different than what is used locally.

    :param app: the WSGI application
    '''
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        script_name = environ.get('HTTP_X_SCRIPT_NAME', '')
        if script_name:
            environ['SCRIPT_NAME'] = script_name
            path_info = environ['PATH_INFO']
            if path_info.startswith(script_name):
                environ['PATH_INFO'] = path_info[len(script_name):]

        server = environ.get('HTTP_X_FORWARDED_HOST', '')
        if server:
            environ['HTTP_HOST'] = server

        scheme = environ.get('HTTP_X_SCHEME', '')
        if scheme:
            environ['wsgi.url_scheme'] = scheme
        return self.app(environ, start_response)



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>
# Written by Matt Prahl <mprahl@redhat.com>
from os import path, environ

import pytest
import mock
from mock import patch, PropertyMock, Mock
import kerberos
import ldap3
from werkzeug.exceptions import Unauthorized as FlaskUnauthorized

import module_build_service.auth
import module_build_service.errors
import module_build_service.config as mbs_config


class TestAuthModule:
    def test_get_user_no_token(self):
        base_dir = path.abspath(path.dirname(__file__))
        client_secrets = path.join(base_dir, "client_secrets.json")
        with patch.dict('module_build_service.app.config', {'OIDC_CLIENT_SECRETS': client_secrets,
                                                            'OIDC_REQUIRED_SCOPE': 'mbs-scope'}):
            request = mock.MagicMock()
            request.cookies.return_value = {}

            with pytest.raises(module_build_service.errors.Unauthorized) as cm:
                module_build_service.auth.get_user(request)
                assert str(cm.value) == "No 'authorization' header found."

    @patch('module_build_service.auth._get_token_info')
    @patch('module_build_service.auth._get_user_info')
    def test_get_user_failure(self, get_user_info, get_token_info):
        base_dir = path.abspath(path.dirname(__file__))
        client_secrets = path.join(base_dir, "client_secrets.json")
        with patch.dict('module_build_service.app.config', {'OIDC_CLIENT_SECRETS': client_secrets,
                                                            'OIDC_REQUIRED_SCOPE': 'mbs-scope'}):
            # https://www.youtube.com/watch?v=G-LtddOgUCE
            name = "Joey Jo Jo Junior Shabadoo"
            mocked_get_token_info = {"active": False, "username": name,
                                     "scope": ("openid https://id.fedoraproject.org/scope/groups"
                                               " mbs-scope")}
            get_token_info.return_value = mocked_get_token_info

            get_user_info.return_value = {"groups": ["group"]}

            headers = {"authorization": "Bearer foobar"}
            request = mock.MagicMock()
            request.headers.return_value = mock.MagicMock(spec_set=dict)
            request.headers.__getitem__.side_effect = headers.__getitem__
            request.headers.__setitem__.side_effect = headers.__setitem__
            request.headers.__contains__.side_effect = headers.__contains__

            with pytest.raises(module_build_service.errors.Unauthorized) as cm:
                module_build_service.auth.get_user(request)
                assert str(cm.value) == "OIDC token invalid or expired."

    @patch('module_build_service.auth._get_token_info')
    @patch('module_build_service.auth._get_user_info')
    def test_get_user_good(self, get_user_info, get_token_info):
        base_dir = path.abspath(path.dirname(__file__))
        client_secrets = path.join(base_dir, "client_secrets.json")
        with patch.dict('module_build_service.app.config', {'OIDC_CLIENT_SECRETS': client_secrets,
                                                            'OIDC_REQUIRED_SCOPE': 'mbs-scope'}):
            # https://www.youtube.com/watch?v=G-LtddOgUCE
            name = "Joey Jo Jo Junior Shabadoo"
            mocked_get_token_info = {"active": True, "username": name,
                                     "scope": ("openid https://id.fedoraproject.org/scope/groups"
                                               " mbs-scope")}
            get_token_info.return_value = mocked_get_token_info

            get_user_info.return_value = {"groups": ["group"]}

            headers = {"authorization": "Bearer foobar"}
            request = mock.MagicMock()
            request.headers.return_value = mock.MagicMock(spec_set=dict)
            request.headers.__getitem__.side_effect = headers.__getitem__
            request.headers.__setitem__.side_effect = headers.__setitem__
            request.headers.__contains__.side_effect = headers.__contains__

            username, groups = module_build_service.auth.get_user(request)
            assert username == name
            assert groups == set(get_user_info.return_value["groups"])

    @patch.object(mbs_config.Config, 'no_auth', new_callable=PropertyMock, return_value=True)
    def test_disable_authentication(self, conf_no_auth):
        request = mock.MagicMock()
        username, groups = module_build_service.auth.get_user(request)
        assert username == "anonymous"
        assert groups == {"packager"}

    @patch('module_build_service.auth.client_secrets', None)
    def test_misconfiguring_oidc_client_secrets_should_be_failed(self):
        request = mock.MagicMock()
        with pytest.raises(module_build_service.errors.Forbidden) as cm:
            module_build_service.auth.get_user(request)
            assert str(cm.value) == "OIDC_CLIENT_SECRETS must be set in server config."

    @patch('module_build_service.auth._get_token_info')
    @patch('module_build_service.auth._get_user_info')
    def test_get_required_scope_not_present(self, get_user_info, get_token_info):
        base_dir = path.abspath(path.dirname(__file__))
        client_secrets = path.join(base_dir, "client_secrets.json")
        with patch.dict('module_build_service.app.config', {'OIDC_CLIENT_SECRETS': client_secrets,
                                                            'OIDC_REQUIRED_SCOPE': 'mbs-scope'}):
            # https://www.youtube.com/watch?v=G-LtddOgUCE
            name = "Joey Jo Jo Junior Shabadoo"
            mocked_get_token_info = {"active": True,
                                     "username": name,
                                     "scope": "openid https://id.fedoraproject.org/scope/groups"}
            get_token_info.return_value = mocked_get_token_info

            get_user_info.return_value = {"groups": ["group"]}

            headers = {"authorization": "Bearer foobar"}
            request = mock.MagicMock()
            request.headers.return_value = mock.MagicMock(spec_set=dict)
            request.headers.__getitem__.side_effect = headers.__getitem__
            request.headers.__setitem__.side_effect = headers.__setitem__
            request.headers.__contains__.side_effect = headers.__contains__

            with pytest.raises(module_build_service.errors.Unauthorized) as cm:
                module_build_service.auth.get_user(request)
                assert str(cm.value) == ("Required OIDC scope 'mbs-scope' not present: "
                                         "['openid', 'https://id.fedoraproject.org/scope/groups']")

    @patch('module_build_service.auth._get_token_info')
    @patch('module_build_service.auth._get_user_info')
    def test_get_required_scope_not_set_in_cfg(self, get_user_info, get_token_info):
        base_dir = path.abspath(path.dirname(__file__))
        client_secrets = path.join(base_dir, "client_secrets.json")
        with patch.dict('module_build_service.app.config', {'OIDC_CLIENT_SECRETS': client_secrets}):
            # https://www.youtube.com/watch?v=G-LtddOgUCE
            name = "Joey Jo Jo Junior Shabadoo"
            mocked_get_token_info = {"active": True,
                                     "username": name,
                                     "scope": "openid https://id.fedoraproject.org/scope/groups"}
            get_token_info.return_value = mocked_get_token_info

            get_user_info.return_value = {"groups": ["group"]}

            headers = {"authorization": "Bearer foobar"}
            request = mock.MagicMock()
            request.headers.return_value = mock.MagicMock(spec_set=dict)
            request.headers.__getitem__.side_effect = headers.__getitem__
            request.headers.__setitem__.side_effect = headers.__setitem__
            request.headers.__contains__.side_effect = headers.__contains__

            with pytest.raises(module_build_service.errors.Forbidden) as cm:
                module_build_service.auth.get_user(request)
                assert str(cm.value) == "OIDC_REQUIRED_SCOPE must be set in server config."


class KerberosMockConfig(object):
    def __init__(self, uri='ldaps://test.example.local:636', dn='ou=groups,dc=domain,dc=local',
                 kt='/path/to/keytab', host='mbs.domain.local'):
        """
        :param uri: a string overriding config.ldap_uri
        :param dn: a string overriding config.ldap_groups_dn
        :param kt: a string overriding config.kerberos_keytab
        :param host: a string overriding config.kerberos_http_host
        """
        self.uri = uri
        self.dn = dn
        self.kt = kt
        self.host = host

    def __enter__(self):
        self.auth_method_p = patch.object(
            mbs_config.Config, 'auth_method', new_callable=PropertyMock)
        mocked_auth_method = self.auth_method_p.start()
        mocked_auth_method.return_value = 'kerberos'

        self.ldap_uri_p = patch.object(
            mbs_config.Config, 'ldap_uri', new_callable=PropertyMock)
        mocked_ldap_uri = self.ldap_uri_p.start()
        mocked_ldap_uri.return_value = self.uri

        self.ldap_dn_p = patch.object(
            mbs_config.Config, 'ldap_groups_dn', new_callable=PropertyMock)
        mocked_ldap_dn = self.ldap_dn_p.start()
        mocked_ldap_dn.return_value = self.dn

        self.kerberos_keytab_p = patch.object(
            mbs_config.Config, 'kerberos_keytab', new_callable=PropertyMock)
        mocked_kerberos_keytab = self.kerberos_keytab_p.start()
        mocked_kerberos_keytab.return_value = self.kt

        self.kerberos_http_host_p = patch.object(
            mbs_config.Config, 'kerberos_http_host', new_callable=PropertyMock)
        mocked_kerberos_http_host = self.kerberos_http_host_p.start()
        mocked_kerberos_http_host.return_value = self.host

    def __exit__(self, *args):
        self.auth_method_p.stop()
        self.ldap_uri_p.stop()
        self.ldap_dn_p.stop()
        self.kerberos_keytab_p.stop()
        self.kerberos_http_host_p.stop()


class TestAuthModuleKerberos:
    @patch('kerberos.authGSSServerInit', return_value=(kerberos.AUTH_GSS_COMPLETE, object()))
    @patch('kerberos.authGSSServerStep', return_value=kerberos.AUTH_GSS_COMPLETE)
    @patch('kerberos.authGSSServerResponse', return_value='STOKEN')
    @patch('kerberos.authGSSServerUserName', return_value='mprahl@EXAMPLE.ORG')
    @patch('kerberos.authGSSServerClean')
    @patch('kerberos.getServerPrincipalDetails')
    @patch.dict('os.environ')
    @patch('module_build_service.auth.stack')
    def test_get_user_kerberos(self, stack, principal, clean, name, response,
                               step, init):
        """
        Test that authentication works with Kerberos and LDAP
        """
        mock_top = Mock()
        stack.return_value = mock_top

        headers = {'Authorization': 'foobar'}
        request = mock.MagicMock()
        request.headers.return_value = mock.MagicMock(spec_set=dict)
        request.headers.__getitem__.side_effect = headers.__getitem__
        request.headers.__setitem__.side_effect = headers.__setitem__
        request.headers.__contains__.side_effect = headers.__contains__

        # Create the mock LDAP instance
        server = ldap3.Server('ldaps://test.domain.local')
        connection = ldap3.Connection(server, client_strategy=ldap3.MOCK_SYNC)
        base_dn = 'dc=domain,dc=local'
        factory_group_attrs = {
            'objectClass': ['top', 'posixGroup'],
            'memberUid': ['mprahl', 'tbrady'],
            'gidNumber': 1234,
            'cn': ['factory2-devs']
        }
        devs_group_attrs = {
            'objectClass': ['top', 'posixGroup'],
            'memberUid': ['mprahl', 'mikeb'],
            'gidNumber': 1235,
            'cn': ['devs']
        }
        athletes_group_attrs = {
            'objectClass': ['top', 'posixGroup'],
            'memberUid': ['tbrady', 'rgronkowski'],
            'gidNumber': 1236,
            'cn': ['athletes']
        }
        mprahl_attrs = {
            'memberOf': ['cn=Employee,ou=groups,{0}'.format(base_dn)],
            'uid': ['mprahl'],
            'cn': ['mprahl'],
            'objectClass': ['top', 'person']
        }
        connection.strategy.add_entry('cn=factory2-devs,ou=groups,{0}'.format(base_dn),
                                      factory_group_attrs)
        connection.strategy.add_entry('cn=athletes,ou=groups,{0}'.format(base_dn),
                                      athletes_group_attrs)
        connection.strategy.add_entry('cn=devs,ou=groups,{0}'.format(base_dn), devs_group_attrs)
        connection.strategy.add_entry('cn=mprahl,ou=users,{0}'.format(base_dn), mprahl_attrs)

        groups = {'devs', 'factory2-devs'}
        with patch('ldap3.Connection') as mock_ldap_con, KerberosMockConfig():
            mock_ldap_con.return_value = connection
            assert module_build_service.auth.get_user_kerberos(request) == ('mprahl', groups)

    def test_auth_header_not_set(self):
        """
        Test that an Unauthorized exception is returned when there is no authorization header
        set.
        """
        headers = {}
        request = mock.MagicMock()
        request.headers.return_value = mock.MagicMock(spec_set=dict)
        request.headers.__getitem__.side_effect = headers.__getitem__
        request.headers.__setitem__.side_effect = headers.__setitem__
        request.headers.__contains__.side_effect = headers.__contains__

        with KerberosMockConfig():
            try:
                module_build_service.auth.get_user_kerberos(request)
                assert False, 'Unauthorized error not raised'
            except FlaskUnauthorized as error:
                assert error.response.www_authenticate.to_header().strip() == 'Negotiate'
                assert error.response.status == '401 UNAUTHORIZED'

    @patch.dict(environ)
    def test_keytab_not_set(self):
        """
        Test that authentication fails when the keytab is not set
        """
        if 'KRB5_KTNAME' in environ:
            del environ['KRB5_KTNAME']

        headers = {'Authorization': 'foobar'}
        request = mock.MagicMock()
        request.headers.return_value = mock.MagicMock(spec_set=dict)
        request.headers.__getitem__.side_effect = headers.__getitem__
        request.headers.__setitem__.side_effect = headers.__setitem__
        request.headers.__contains__.side_effect = headers.__contains__

        with KerberosMockConfig(kt=''):
            try:
                module_build_service.auth.get_user_kerberos(request)
                assert False, 'Unauthorized error not raised'
            except module_build_service.errors.Unauthorized as error:
                assert str(error) == ('Kerberos: set the config value of "KERBEROS_KEYTAB" '
                                      'or the environment variable "KRB5_KTNAME" to your '
                                      'keytab file')

    # Set the return value to something not 0 (continue) or 1 (complete)
    @patch('kerberos.authGSSServerInit', return_value=(100, object()))
    @patch('kerberos.authGSSServerStep', return_value=kerberos.AUTH_GSS_COMPLETE)
    @patch('kerberos.authGSSServerResponse', return_value='STOKEN')
    @patch('kerberos.authGSSServerUserName', return_value='mprahl@EXAMPLE.ORG')
    @patch('kerberos.authGSSServerClean')
    @patch('kerberos.getServerPrincipalDetails')
    @patch.dict('os.environ')
    @patch('module_build_service.auth.stack')
    def test_get_user_kerberos_invalid_ticket(self, stack, principal, clean, name, response,
                                              step, init):
        """
        Test that authentication fails with an invalid Kerberos ticket
        """
        mock_top = Mock()
        stack.return_value = mock_top

        headers = {'Authorization': 'foobar'}
        request = mock.MagicMock()
        request.headers.return_value = mock.MagicMock(spec_set=dict)
        request.headers.__getitem__.side_effect = headers.__getitem__
        request.headers.__setitem__.side_effect = headers.__setitem__
        request.headers.__contains__.side_effect = headers.__contains__

        with KerberosMockConfig():
            try:
                module_build_service.auth.get_user_kerberos(request)
                assert False, 'Forbidden error not raised'
            except module_build_service.errors.Forbidden as error:
                assert str(error) == ('Invalid Kerberos ticket')



"""Adds the owner, time_completed, time_modified, and time_submitted columns to the module_builds table

Revision ID: 1a44272e8b4c
Revises: a7a553e5ca1d
Create Date: 2016-08-17 17:00:31.126429

"""

# revision identifiers, used by Alembic.
revision = '1a44272e8b4c'
down_revision = 'a7a553e5ca1d'

from alembic import op
import sqlalchemy as sa
from datetime import datetime

epoch = datetime.utcfromtimestamp(0).strftime('%Y-%m-%d %H:%M:%S')


def upgrade():
    op.add_column('module_builds', sa.Column('owner', sa.String(), server_default='Unknown User', nullable=False))
    op.add_column('module_builds', sa.Column('time_completed', sa.DateTime(), nullable=True, server_default=epoch))
    op.add_column('module_builds', sa.Column('time_modified', sa.DateTime(), nullable=True, server_default=epoch))
    op.add_column('module_builds', sa.Column('time_submitted', sa.DateTime(), nullable=False, server_default=epoch))

    # Remove migration-only defaults. Using batch_alter_table() recreates the table instead of using ALTER COLUMN
    # on simplistic DB engines. Thanks SQLite!
    with op.batch_alter_table('module_builds') as b:
        b.alter_column('owner', server_default=None)
        b.alter_column('time_completed', server_default=None)
        b.alter_column('time_modified', server_default=None)
        b.alter_column('time_submitted', server_default=None)


def downgrade():
    # Thanks again!
    with op.batch_alter_table('module_builds') as b:
        b.drop_column('time_submitted')
        b.drop_column('time_modified')
        b.drop_column('time_completed')
        b.drop_column('owner')
